/* bdb-chol.cp.c */

/* This program is the node processor software for a parallel version
of the Shur complement Choleski factorization algorithm for
block-diagonal-bordered form sparse symmetric matrices.  */

/* special considerations for this implementation: 

1) Matrices are in symmetric block-diagonal-bordered form

2) any number of processors can be specified up to an including the
total number of processors on the CM-5 

3) it is assumed that ordering and load balancing are performed a
priori

4) the last block is considered dense, and a dense blocked symmetric
Choleski algorithm is used to factor this portion of the matrix

*/

/* Naming convention used throughout this software

i - row number
j - column number
k - number along the diagonal

Lij   - any value in a strictly lower diagonal matrix block
Lkk   - value of the kth element on the lower diagonal
        stored in the location diag[k].Value
Lix   - row in a strictly lower diagonal matrix block 
        (ith row contributing to update of Lij)
Ljx   - row in a strictly lower diagonal matrix block 
        (jth row contributing to update of Lij)
        Because matrix is symmetric, this row would 
        correspond to the column Uxj in L^T
Lxk   - column in a strictly lower diagonal matrix block
        only accessed when updating a column

LBij  - any value in a lower  matrix border
LBix  - row in a lower matrix border - (ith row contributing 
        to update of LBij)
LBjx  - row in a matrix  border (jth row contributing 
        to update of LBij)   Because matrix is symmetric, this row would 
        correspond to the border column UBxj in L^T
LBxk  - column in a lower matrix border
        only accessed when updating a column

locLijLkj    - function to find the implicit storage locations 
               of Lij and Lkj, where i and k are given, and 
               (Lij != 0.) && (Lkj != 0.) && (j < k)
locLBijLkj   - function to find the implicit storage locations 
               of LBij and Ujk, where i and k are given, and 
               (LBij != 0.) && (Ujk != 0.) && (j < k)
LBixLBkx    -  function to find the sum of the products of values
               in partial row LBix and partial row LBkx
locLBixLBkx - function to find the implicit storage locations 
               of LBij and LBkj, where i and k are given, and 
               (LBij != 0.) && (LBkj != 0.)
*/
/*---------------------------------------------------------*/
#include "AIXtranslate.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <malloc.h>
#include <math.h>
/* #include <cm/cmmd.h> */
/* #include "tbx_timer.c"  */
#include "bdb_struct_par_chol.h"
#include "bdb_chol_MSG_NUM.h"
/*---------------------------------------------------------*/
#define YES        1
#define NO         0
/*---------------------------------------------------------*/
typedef struct Node_Timer{int pn_id;
                          double elapsed;
                          double busy;
                          double idle;} node_timer;
node_timer node_tim;

typedef struct CMMD_Timer{double min;
                          double ave;
                          double max;} CMMD_Timer;
CMMD_Timer *tSTAT;  /* tSTAT[] */
CMMD_Timer tFAC,tFR,tBS,tTOT;
/*---------------------------------------------------------*/
typedef struct BBuf{int nprocs;
                    int n;
                    int nblocks;
                    int nblockrow;
                    int nlborder; 
                    int nborderrow;
                    int ln;
                    int nb;} BBuf;
BBuf bbuf;
/*---------------------------------------------------------*/
Proc        *proc;       /* proc[]      */
/*---------------------------------------------------------*/
Block       *blocks;     /* blocks[]     */
Block       *pblocks;    /* pblocks[]    */
/*---------------------------------------------------------*/
Diag        *diag;       /* diag[]       */
Diag        *pdiag;      /* pdiag[]      */
/*---------------------------------------------------------*/
BlockRow    *blockrow;   /* blockrow[]   */
BlockRow    *pblockrow;  /* pblockrow[]  */
/*---------------------------------------------------------*/
LowerBorder *lborder;    /* lborder[]    */
LowerBorder *plborder;   /* plborder[]   */
/*---------------------------------------------------------*/
BorderRow   *borderrow;  /* borderrow[]  */
BorderRow   *pborderrow; /* pborderrow[] */
/*---------------------------------------------------------*/
double         *b;         /* b[]         */
double         *pb;        /* pb[]        */
/*---------------------------------------------------------*/
int            n;          /* dimension of block-diagonal block */
int            nblocks;    /* # of blocks            */
int            nblockrow;  /* dimension of blockrow  */
int            nlborder;   /* dimension of lborder   */
int            nborderrow; /* dimension of borderrow */
/*---------------------------------------------------------*/
int            nprocs;
int            bytesreceived;
int            bytessent;
/*---------------------------------------------------------*/
double *la;  /* la[] stored in column-major order (ie., a FORTRAN matrix) */
double *lb;  /* lb[]  */
int    ln;   /* dimension of the last block */
int    nb;   /* block size */
/*---------------------------------------------------------*/
/* function prototypes */
void   InputData();
int    min();

main(argc,argv)
  int argc;
  char **argv;
{
  int i,j,l,bufsize,nodesperproc;
#include "AIX.h"

  printf("\n Executing host program... (pid=%d)\n", getpid());

  if (argc < 4)
   {
    printf("Usage: (1) input_file_name_A_matrix \n");
    printf("       (2) input_file_name_b_vector \n");
    printf("       (3) block size \n");
    exit(0);
   } /* end if */

  CMMD_enable();

/* input the data from the file - allocate memory as required */
  printf("input file for the A matrix  = |%s|\n",argv[1]);
  printf("input file for b vector  = |%s|\n",argv[2]);
  sscanf(argv[3],"%d",&nb);
  printf("block size = |%d| \n",nb);
  fflush(stdout);

printf("InputData() \n");
fflush(stdout);
  InputData(argv[1],argv[2]);

printf("FactorIndependentBlocks() \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FIB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[0].min) tSTAT[0].min = node_tim.busy;
    if (node_tim.busy > tSTAT[0].max) tSTAT[0].max = node_tim.busy;
    tSTAT[0].ave += node_tim.busy;
   } /* end for i */
  tSTAT[0].ave /= (double) nprocs;
  tFAC.min += tSTAT[0].min;
  tFAC.max += tSTAT[0].max;
  tFAC.ave += tSTAT[0].ave;
  tTOT.min += tSTAT[0].min;
  tTOT.max += tSTAT[0].max;
  tTOT.ave += tSTAT[0].ave;

printf("UpdateLastBlock() \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_ULB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[1].min) tSTAT[1].min = node_tim.busy;
    if (node_tim.busy > tSTAT[1].max) tSTAT[1].max = node_tim.busy;
    tSTAT[1].ave += node_tim.busy;
  } /* end for i */
  tSTAT[1].ave /= (double) nprocs;
  tFAC.min += tSTAT[1].min;
  tFAC.max += tSTAT[1].max;
  tFAC.ave += tSTAT[1].ave;
  tTOT.min += tSTAT[1].min;
  tTOT.max += tSTAT[1].max;
  tTOT.ave += tSTAT[1].ave;

printf("FactorLastBlock() \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FLB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[2].min) tSTAT[2].min = node_tim.busy;
    if (node_tim.busy > tSTAT[2].max) tSTAT[2].max = node_tim.busy;
    tSTAT[2].ave += node_tim.busy;
   } /* end for i */
  tSTAT[2].ave /= (double) nprocs;
  tFAC.min += tSTAT[2].min;
  tFAC.max += tSTAT[2].max;
  tFAC.ave += tSTAT[2].ave;
  tTOT.min += tSTAT[2].min;
  tTOT.max += tSTAT[2].max;
  tTOT.ave += tSTAT[2].ave;

printf("ForwardReduce - independent blocks \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FRIB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[3].min) tSTAT[3].min = node_tim.busy;
    if (node_tim.busy > tSTAT[3].max) tSTAT[3].max = node_tim.busy;
    tSTAT[3].ave += node_tim.busy;
   } /* end for i */
  tSTAT[3].ave /= (double) nprocs;
  tFR.min += tSTAT[3].min;
  tFR.max += tSTAT[3].max;
  tFR.ave += tSTAT[3].ave;
  tTOT.min += tSTAT[3].min;
  tTOT.max += tSTAT[3].max;
  tTOT.ave += tSTAT[3].ave;

printf("ForwardReduce - lower border \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FRLB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[4].min) tSTAT[4].min = node_tim.busy;
    if (node_tim.busy > tSTAT[4].max) tSTAT[4].max = node_tim.busy;
    tSTAT[4].ave += node_tim.busy;
   } /* end for i */
  tSTAT[4].ave /= (double) nprocs;
  tFR.min += tSTAT[4].min;
  tFR.max += tSTAT[4].max;
  tFR.ave += tSTAT[4].ave;
  tTOT.min += tSTAT[4].min;
  tTOT.max += tSTAT[4].max;
  tTOT.ave += tSTAT[4].ave;

printf("ForwardReduce - last block \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FR,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[5].min) tSTAT[5].min = node_tim.busy;
    if (node_tim.busy > tSTAT[5].max) tSTAT[5].max = node_tim.busy;
    tSTAT[5].ave += node_tim.busy;
   } /* end for i */
  tSTAT[5].ave /= (double) nprocs;
  tFR.min += tSTAT[5].min;
  tFR.max += tSTAT[5].max;
  tFR.ave += tSTAT[5].ave;
  tTOT.min += tSTAT[5].min;
  tTOT.max += tSTAT[5].max;
  tTOT.ave += tSTAT[5].ave;

printf("BackSubLastBlock() - last block \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_BSLB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[6].min) tSTAT[6].min = node_tim.busy;
    if (node_tim.busy > tSTAT[6].max) tSTAT[6].max = node_tim.busy;
    tSTAT[6].ave += node_tim.busy;
   } /* end for i */
  tSTAT[6].ave /= (double) nprocs;
  tBS.min += tSTAT[6].min;
  tBS.max += tSTAT[6].max;
  tBS.ave += tSTAT[6].ave;
  tTOT.min += tSTAT[6].min;
  tTOT.max += tSTAT[6].max;
  tTOT.ave += tSTAT[6].ave;

printf("BackwardSubstitution() - independent blocks \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_BSIB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[7].min) tSTAT[7].min = node_tim.busy;
    if (node_tim.busy > tSTAT[7].max) tSTAT[7].max = node_tim.busy;
    tSTAT[7].ave += node_tim.busy;
   } /* end for i */  
  tSTAT[7].ave /= (double) nprocs;
  tBS.min += tSTAT[7].min;
  tBS.max += tSTAT[7].max;
  tBS.ave += tSTAT[7].ave;
  tTOT.min += tSTAT[7].min;
  tTOT.max += tSTAT[7].max;
  tTOT.ave += tSTAT[7].ave;

  printf("\nNumber of blocks = |%d| \n",(nblocks+1));
  printf("Total number of nonzeros = |%d| \n",
           (n+nblockrow+nborderrow+(ln*(ln-1)/2)));
  printf("Fraction of nonzeros in the matrix = |%lf| \n\n",
          ((double)(n+nblockrow+nborderrow+(ln*(ln-1)/2)) / (double)((n+ln)*(n+ln))) );
  fflush(stdout);

  printf("\nTIMING STATISTICS \n\n");
  printf("\nFactor \t\t  tFAC.min=%lf \t tFAC.ave=%lf \t tFAC.max=%lf\n",
                tFAC.min,tFAC.ave,tFAC.max);
  printf("Forward Reduce\t  tFR.min= %lf \t tFR.ave= %lf \t tFR.max= %lf\n",
                tFR.min,tFR.ave,tFR.max);
  printf("Back Sub \t  tBS.min= %lf \t tBS.ave= %lf \t tBS.max= %lf\n",
                tBS.min,tBS.ave,tBS.max);
  printf("Total \t\t  tTOT.min=%lf \t tTOT.ave=%lf \t tTOT.max=%lf\n\n",
                tTOT.min,tTOT.ave,tTOT.max);
  for (i=0;i<8;i++)
    {
     printf("\t tSTAT[%d].min=%lf \t tSTAT[%d].ave=%lf \t tSTAT[%d].max=%lf\n",
               i,tSTAT[i].min,i,tSTAT[i].ave,i,tSTAT[i].max);
    } /* end for i */

} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void InputData(fnameA,fnameb)
  char *fnameA,*fnameb;
{
/* input file format for block-bordered diagonal A matrix

(nblocks - dimension of blocks[])
blocks[1].BlockID  blocks[1].BlockSz blocks[1].DiagPtr blocks[1].LBPtr
.
.
.
blocks[nblocks].BlockID blocks[nblocks].BlockSz blocks[nblocks].DiagPtr blocks[nblocks].LBPtr

(n - dimension of diag[] -  i.e., the matrix dimension)
diag[1].Value diag[1].PtrLix diag[1].NextLxk diag[1].NextLBxk
.
.
.
diag[n].Value diag[n].PtrLix diag[n].NextLxk diag[n].NextLBxk

(nblockrow - dimension of blockrow[])
blockrow[1].Value blockrow[1].i blockrow[1].j blockrow[1].NextLxk
.
.
.
blockrow[nblockrow].Value blockrow[nblockrow].i blockrow[nblockrow].j blockrow[nblockrow].NextLxk

(nlborder - dimension of lborder[])
lborder[1].i        lborder[1].PtrLBix
.
.
.
lborder[nlborder].i lborder[nlborder].PtrLBix

(nborderrow - dimension of borderrow[])
borderrow[1].Value borderrow[1].i borderrow[1].j  borderrow[1].NextLBxk
.
.
.
borderrow[nborderrow].Value borderrow[nborderrow].i borderrow[nborderrow].j borderrow[nborderrow].NextLBxk

nl nlb
1 1 a(1,1)
1 2 a(1,2)
.
.
.
i j a(i,j)
.
.
.

EOF_input_A_matrix

Input file format for the b vector

b[1]
.
.
.
b[n]

EOF_input_b_vector

*/
  int c,i,j,k,l,m,iproc,jb,bufsize,nlb;
  int np;          /* number of diagonal elements distributed to a processor */
  int npblocks;    /* # of blocks distributed to a processor */
  int npblockrow;  /* dimension of blockrow for a processor */
  int nplborder;   /* dimension of lborder for a processor */
  int npborderrow; /* dimension of borderrow for a processor */
  int lastblockstart,lastblockend;
  double aij;

  char *EOFCheck;
  FILE *fd, *fopen();
#include "AIX.h"

  EOFCheck = (char *) calloc(80,sizeof(char));
  if (EOFCheck == NULL)
   {
    printf("can't allocate memory for EOFCheck \n");
    exit(1);
   } /* end if */

  fd = fopen(fnameA,"r");
  if (fd == NULL)
   {
    printf("can't open file %s\n",fnameA);
    exit(1);
   } /* end if */

/* (I-0) read in the processor identifier array of structures */
  fscanf(fd,"%d",&nprocs);
printf("nprocs=|%d| \n",nprocs);
fflush(stdout);
  proc = (Proc *) calloc((nprocs+1),sizeof(Proc));
  if (proc == NULL)
   {
    printf("can't allocate memory for proc[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nprocs; i++)
   {
    fscanf(fd,"%d %d %d %d %d %d %d",&c,&c,&(proc[i].BlockPtr),&(proc[i].DiagPtr),
           &(proc[i].BlockRowPtr),&(proc[i].LowerBorderPtr),&(proc[i].BorderRowPtr));
   } /* end for i */


/* (I-1) read in the block identifier array of structures */
  fscanf(fd,"%d",&nblocks);
printf("nblocks=|%d| \n",nblocks);
fflush(stdout);
  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblocks; i++)
   {
    fscanf(fd,"%d %d %d %d %d %d",&c,&(blocks[i].BlockID),&(blocks[i].BlockSz),
            &(blocks[i].ProcNum),&(blocks[i].DiagPtr),&(blocks[i].LBPtr) );
   } /* end for i */

/* (I-2) read in the matrix diagonal array of structures */
  fscanf(fd,"%d",&n);
printf("n=|%d| \n",n);
fflush(stdout);
  diag = (Diag *) calloc((n+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=n; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,&(diag[i].Value),&(diag[i].PtrLix),
            &(diag[i].NextLxk),&(diag[i].NextLBxk) );
   } /* end for i */

/* (I-3) read in the block row storage array of structures */
  fscanf(fd,"%d",&nblockrow);
printf("nblockrow=|%d| \n",nblockrow);
fflush(stdout);
  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblockrow; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,&(blockrow[i].Value),&(blockrow[i].i),
            &(blockrow[i].j),&(blockrow[i].NextLxk) );
   } /* end for i */

/* (I-5) read in the lower border row storage array of structures */
  fscanf(fd,"%d",&nlborder);
printf("nlborder=|%d| \n",nlborder);
fflush(stdout);
  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nlborder; i++)
   {
    fscanf(fd,"%d %d %d",&c,&(lborder[i].i),&(lborder[i].PtrLBix) );
   } /* end for i */

/* (I-7) read in the border row storage array of structures */
  fscanf(fd,"%d",&nborderrow);
printf("nborderrow=|%d| \n",nborderrow);
fflush(stdout);
  borderrow = (BorderRow *) calloc((nborderrow+1),sizeof(BorderRow));
  if (borderrow == NULL)
   {
    printf("can't allocate memory for borderrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nborderrow; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,&(borderrow[i].Value),&(borderrow[i].i),
            &(borderrow[i].j),&(borderrow[i].NextLBxk) );
   } /* end for i */

/* (I-9) read in the dense matrix */
  fscanf(fd,"%d %d",&ln,&nlb);
printf("ln=|%d| - nlb=|%d| \n",ln,nlb);
fflush(stdout);
  la = (double *) calloc((ln*ln),sizeof(double));
  if (la == NULL)
   {
    printf("can't allocate memory for la[] \n");
    exit(1);
   } /* end if */
  for (l=1; l<=nlb; l++)
   {
    fscanf(fd,"%d %d %d %lf",&c,&i,&j,&aij);
    i -= (n+1);
    j -= (n+1);
    k = ((j*ln)+i);
/*
if (k < 512) {
  printf("(%d) --- a(%d,%d)=%lf \n",k,i,j,aij);
  fflush(stdout); }
*/
    la[k] = aij;
   } /* end for l */

/* (II-0) load the block structure buffers for the independent blocks */
  bbuf.nprocs = nprocs;
  bbuf.n          = n;
  bbuf.nblocks    = nblocks;
  bbuf.nblockrow  = nblockrow;
  bbuf.nlborder   = nlborder;
  bbuf.nborderrow = nborderrow;
  bbuf.ln         = ln;
  bbuf.nb         = nb;
  CMMD_bc_from_host(&bbuf,(8*sizeof(int)));

/* (II-1) load the block structure buffers for the independent blocks */
printf("load the block structure buffers for the independent blocks npblocks=%d\n",
              npblocks);
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockPtr;
    m = proc[i+1].BlockPtr;
    npblocks = (m-l);
    CMMD_send(i,SEND_NPBLOCKS,&npblocks,sizeof(int));
    pblocks = (Block *) calloc((npblocks+1),sizeof(Block));
    if (pblocks == NULL)
     {
      printf("can't allocate memory for pblocks[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblocks[0].BlockID = 0;
    pblocks[0].BlockSz = 0;
    pblocks[0].DiagPtr = blocks[l].DiagPtr;
    pblocks[0].LBPtr   = 0;

printf("\ti=%d - l=%d - m=%d - pblocks[0].DiagPtr=%d \n",i,l,m,pblocks[0].DiagPtr);
fflush(stdout);

    for (j=1; j<=npblocks; j++)
     {
      pblocks[j].BlockID = blocks[j+l].BlockID;
      pblocks[j].BlockSz = blocks[j+l].BlockSz;
      pblocks[j].DiagPtr = blocks[j+l].DiagPtr;
      if (blocks[j+l].LBPtr == 0)
        pblocks[j].LBPtr = 0;
      else
        pblocks[j].LBPtr = blocks[j+l].LBPtr - blocks[l].LBPtr;
     } /* end for j */
/*
printf("send the block structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBLOCKS,pblocks,((npblocks+1)*sizeof(Block)));
    free(pblocks); 
   } /* end for i */

/* (II-2) load the matrix diagonal structure buffers for the independent blocks */
printf("load the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].DiagPtr;
    m = proc[i+1].DiagPtr;
    np = (m-l);
    CMMD_send(i,SEND_NP,&np,sizeof(int));
    pdiag = (Diag *) calloc((np+1),sizeof(Diag));
    if (pdiag == NULL)
     {
      printf("can't allocate memory for pdiag[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pdiag[0].Value    = 0.0;
    pdiag[0].PtrLix   = 0;
    pdiag[0].NextLxk  = 0;
    pdiag[0].NextLBxk = 0;
    for (j=1; j<=np; j++)
     {
      pdiag[j].Value    = diag[j+l].Value;
      pdiag[j].PtrLix   = diag[j+l].PtrLix - proc[i].BlockRowPtr;
      if (diag[j+l].NextLxk == 0)
        pdiag[j].NextLxk = 0;
      else
        pdiag[j].NextLxk  = diag[j+l].NextLxk - proc[i].BlockRowPtr;
      if (diag[j+l].NextLBxk == 0)
        pdiag[j].NextLBxk = 0;
      else
        pdiag[j].NextLBxk = diag[j+l].NextLBxk - proc[i].BorderRowPtr;
     } /* end for j */
/*
printf("send the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PDIAG,pdiag,((np+1)*sizeof(Diag)));
    free(pdiag);
   } /* end for i */

/* (II-3) load the block row storage structure buffers for the independent blocks */
printf("load the block row storage structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockRowPtr;
    m = proc[i+1].BlockRowPtr;
    npblockrow = (m-l);
    CMMD_send(i,SEND_NPBLOCKROW,&npblockrow,sizeof(int));
    pblockrow = (BlockRow *) calloc((npblockrow+1),sizeof(BlockRow));
    if (pblockrow == NULL)
     {
      printf("can't allocate memory for pblockrow[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblockrow[0].Value   = 0.0;
    pblockrow[0].i       = 0;
    pblockrow[0].j       = 0;
    pblockrow[0].NextLxk = 0;
    for (j=1; j<=npblockrow; j++)
     {
      pblockrow[j].Value   = blockrow[j+l].Value;
      pblockrow[j].i       = blockrow[j+l].i;
      pblockrow[j].j       = blockrow[j+l].j;
      if (blockrow[j+l].NextLxk == 0)
        pblockrow[j].NextLxk = 0;
      else
        pblockrow[j].NextLxk = blockrow[j+l].NextLxk - l;
     } /* end for j */
/*
printf("send the block row storage structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBLOCKROW,pblockrow,((npblockrow+1)*sizeof(BlockRow)));
    free(pblockrow);
   } /* end for i */

/* (II-5) load the lower border row storage buffers for the independent blocks */
printf("load the lower border row storage buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].LowerBorderPtr;
    m = proc[i+1].LowerBorderPtr;
    nplborder = (m-l);
    CMMD_send(i,SEND_NPLBORDER,&nplborder,sizeof(int));
    plborder = (LowerBorder *) calloc((nplborder+1),sizeof(LowerBorder));
    if (plborder == NULL)
     {
      printf("can't allocate memory for plborder[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    plborder[0].i       = 0;
    plborder[0].PtrLBix = 0;
    for (j=1; j<=nplborder; j++)
     {
      plborder[j].i       = lborder[j+l].i;
      plborder[j].PtrLBix = lborder[j+l].PtrLBix - proc[i].BorderRowPtr;
     } /* end for j */
/*
printf("send the lower border row storage buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PLBORDER,plborder,((nplborder+1)*sizeof(LowerBorder)));
    free(plborder);
   } /* end for i */

/* (II-7) load the block diagonal structure buffers for the independent blocks */
printf("load the block diagonal structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BorderRowPtr;
    m = proc[i+1].BorderRowPtr;
    npborderrow = (m-l);
    CMMD_send(i,SEND_NPBORDERROW,&npborderrow,sizeof(int));
    pborderrow = (BorderRow *) calloc((npborderrow+1),sizeof(BorderRow));
    if (pborderrow == NULL)
     {
      printf("can't allocate memory for pborderrow[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pborderrow[0].Value    = 0.0;
    pborderrow[0].i        = 0;
    pborderrow[0].j        = 0;
    pborderrow[0].NextLBxk = 0;
    for (j=1; j<=npborderrow; j++)
     {
      pborderrow[j].Value    = borderrow[j+l].Value;
      pborderrow[j].i        = borderrow[j+l].i;
      pborderrow[j].j        = borderrow[j+l].j;
      if (borderrow[j+l].NextLBxk == 0)
       pborderrow[j].NextLBxk = 0;
      else
       pborderrow[j].NextLBxk = borderrow[j+l].NextLBxk - l;
     } /* end for j */
/*
printf("send the block diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBORDERROW,pborderrow,((npborderrow+1)*sizeof(BorderRow)));
    free(pborderrow);
   } /* end for i */
  free(borderrow);
  free(lborder);

/* (II-9) load the dense matrix into the processors */
printf("load the dense matrix into the processors \n");
fflush(stdout);
  iproc = 0;
  for (j=1; j<=ln; j += nb)
   {
    jb = min((ln-j+1),nb);
    bufsize = ((ln*jb)*sizeof(double));
/*
printf("send the dense matrix into the processors \n");
fflush(stdout);
*/
    CMMD_send(iproc,SEND_LA,&(la[((j-1)*ln)]),bufsize);
    iproc = (++iproc)%nprocs;
   } /* end for j */

/* (9) read in a character string to identify EOF */
  fscanf(fd,"%s",EOFCheck);
  printf("Input_A_matrix - |%s| \n",EOFCheck);
  fflush(stdout);

  fclose(fd);

  fd = fopen(fnameb,"r");
  if (fd == NULL)
   {
    printf("can't open file %s\n",fnameb);
    exit(1);
   } /* end if */

/* (10) read in the b vector */
printf("read in the b vector \n");
fflush(stdout);
  b = (double *) calloc((n+1),sizeof(double));
  if (b == NULL)
   {
    printf("can't allocate memory for b[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=n; i++)
   {
    fscanf(fd,"%lf",&(b[i]));
   } /* end for i */
/* load the b vector buffers for the independent blocks */
printf("load the b vector buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].DiagPtr;
    m = proc[i+1].DiagPtr;
    np = (m-l);

printf("\tl=%d - m=%d - np=%d \n",l,m,np);
fflush(stdout);

    pb = (double *) calloc((np+1),sizeof(double));
    if (pb == NULL)
     {
      printf("can't allocate memory for pb[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    for (j=1; j<=np; j++)
     {
      pb[j] = b[j+l];
     } /* end for j */
/*
printf("load the b vector buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PB,pb,((np+1)*sizeof(double)));
    free(pb);
   } /* end for i */

/* (IV) read in the lb vector */
  lb = (double *) calloc(ln,sizeof(double));
  if (lb == NULL)
   {
    printf("can't allocate memory for lb[] \n");
    exit(1);
   } /* end if */
  for (i=0; i<ln; i++)
   {
    fscanf(fd,"%lf",&(lb[i]));
   } /* end for i */

/* load the lb vector onto the processors */
printf("load the lb vector onto the processors \n");
fflush(stdout);
  iproc = 0;
  for (j=1; j<=ln; j += nb)
   {
    jb = min((ln-j+1),nb);
    bufsize = (jb*sizeof(double));
/*
printf("send the lb vector onto the processors iproc=%d \n",iproc);
fflush(stdout);
*/
    CMMD_send(iproc,SEND_LB,&(lb[(j-1)]),bufsize);
/*
printf("send the lb vector onto the processors \n");
fflush(stdout);
*/
    iproc = (++iproc)%nprocs;
   } /* end for j */

/* (11) send the row/col locations of the first and last locations in the
   last block */
printf("send the row/col locations of lastblockstart \n");
fflush(stdout);
  lastblockstart = blocks[nblocks].DiagPtr + 1;
  for (iproc=0;iproc<nprocs;iproc++)
    CMMD_send(iproc,SEND_LASTBKST,&lastblockstart,sizeof(int));
/*
printf("send the row/col locations of lastblockend \n");
fflush(stdout);
*/
  lastblockend   = blocks[nblocks].DiagPtr + ln;
  for (iproc=0;iproc<nprocs;iproc++)
    CMMD_send(iproc,SEND_LASTBKND,&lastblockend,sizeof(int));
printf("lastblockend=%d - lastblockstart=%d \n",lastblockend,lastblockstart);
fflush(stdout);


/* (12) read in a character string to identify EOF */
  fscanf(fd,"%s",EOFCheck);
  printf("Input_b_vector - |%s| \n",EOFCheck);
  fflush(stdout);

  fclose(fd);

/* allocate memory for the remaining variables */
  tSTAT = (CMMD_Timer *) calloc(11,sizeof(CMMD_Timer));
  if (tSTAT == NULL)
   {
    printf("can't allocate memory for tSTAT[] \n");
    exit(1);
   } /* end if */

  for (i=0;i<=10;i++)
    {
     tSTAT[i].min = 100000000.0;
     tSTAT[i].ave = 0.0;
     tSTAT[i].max = 0.0;
    } /* end for i */

  tFAC.min = 0.0;
  tFAC.ave = 0.0;
  tFAC.max = 0.0;

  tFR.min = 0.0;
  tFR.ave = 0.0;
  tFR.max = 0.0;

  tBS.min = 0.0;
  tBS.ave = 0.0;
  tBS.max = 0.0;

  tTOT.min = 0.0;
  tTOT.ave = 0.0;
  tTOT.max = 0.0;

} /* end InputData() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
int min(i,j)
 int i,j;
{
  if (i<=j) return i;
   else return j;

} /* end min() */
